/*
-----------------------------------------------------------------------------
This source file is part of OGRE
    (Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/

Copyright (c) 2000-2005 The OGRE Team
Also see acknowledgements in Readme.html

You may use this sample code for anything you like, it is not covered by the
LGPL like the rest of the engine.
-----------------------------------------------------------------------------
*/
/*
-----------------------------------------------------------------------------
Filename:    ExampleFrameListener.h
Description: Defines an example frame listener which responds to frame events.
This frame listener just moves a specified camera around based on
keyboard and mouse movements.
Mouse:    Freelook
W or Up:  Forward
S or Down:Backward
A:        Step left
D:        Step right
             PgUp:     Move upwards
             PgDown:   Move downwards
             F:        Toggle frame rate stats on/off
			 R:        Render mode
             T:        Cycle texture filtering
                       Bilinear, Trilinear, Anisotropic(8)
             P:        Toggle on/off display of camera position / orientation
-----------------------------------------------------------------------------
*/

#ifndef __KdTreeAppListener_H__
#define __KdTreeAppListener_H__

#include "Ogre.h"
#include "OgreKeyEvent.h"
#include "OgreEventListeners.h"
#include "OgreStringConverter.h"
#include "OgreException.h"

using namespace Ogre;

class KdTreeAppListener: public FrameListener, public KeyListener
{
protected:
	int mSceneDetailIndex ;
    Real mMoveSpeed;
    Degree mRotateSpeed;
    Overlay* mDebugOverlay;
	Overlay* mKdTreeOverlay;
	Real mRotationPeriod;
	bool mShowTree;
	bool mFreeMove;

	int mSeqNum;

    void updateStats(void)
    {
        static String currFps = "Current FPS: ";
        static String avgFps = "Average FPS: ";
        static String bestFps = "Best FPS: ";
        static String worstFps = "Worst FPS: ";
        static String tris = "Triangle Count: ";

        // update stats when necessary
        try {
            OverlayElement* guiAvg = OverlayManager::getSingleton().getOverlayElement("Core/AverageFps");
            OverlayElement* guiCurr = OverlayManager::getSingleton().getOverlayElement("Core/CurrFps");
            OverlayElement* guiBest = OverlayManager::getSingleton().getOverlayElement("Core/BestFps");
            OverlayElement* guiWorst = OverlayManager::getSingleton().getOverlayElement("Core/WorstFps");

            const RenderTarget::FrameStats& stats = mWindow->getStatistics();

            guiAvg->setCaption(avgFps + StringConverter::toString(stats.avgFPS));
            guiCurr->setCaption(currFps + StringConverter::toString(stats.lastFPS));
            guiBest->setCaption(bestFps + StringConverter::toString(stats.bestFPS)
                +" "+StringConverter::toString(stats.bestFrameTime)+" ms");
            guiWorst->setCaption(worstFps + StringConverter::toString(stats.worstFPS)
                +" "+StringConverter::toString(stats.worstFrameTime)+" ms");

            OverlayElement* guiTris = OverlayManager::getSingleton().getOverlayElement("Core/NumTris");
            guiTris->setCaption(tris + StringConverter::toString(stats.triangleCount));

            OverlayElement* guiDbg = OverlayManager::getSingleton().getOverlayElement("Core/DebugText");
            guiDbg->setCaption(mWindow->getDebugText());
        }
        catch(...)
        {
            // ignore
        }
    }

public:

    // Constructor takes a RenderWindow because it uses that to determine input context
    KdTreeAppListener(RenderWindow* win, SceneManager* sm, Real rs, Real ms, Real rp,
		bool useBufferedInputKeys = false, bool useBufferedInputMouse = false)
    {
        mDebugOverlay = OverlayManager::getSingleton().getByName("Core/DebugOverlay");
		mKdTreeOverlay = OverlayManager::getSingleton().getByName("KdTree/DebugOverlay");


        mUseBufferedInputKeys = useBufferedInputKeys;
		mUseBufferedInputMouse = useBufferedInputMouse;
		mInputTypeSwitchingOn = mUseBufferedInputKeys || mUseBufferedInputMouse;

		if (mInputTypeSwitchingOn)
		{
            mEventProcessor = new EventProcessor();
			mEventProcessor->initialise(win);
			mEventProcessor->startProcessingEvents();
			mEventProcessor->addKeyListener(this);
			mInputDevice = mEventProcessor->getInputReader();

		}
        else
        {
            mInputDevice = PlatformManager::getSingleton().createInputReader();
            mInputDevice->initialise(win,true, true);
        }

		mFreeMove = false;
		mShowTree = false;
		mCullCamera = false;
		mCullTransVec = Vector3::ZERO;
		mSeqNum = 0;

		mCamera = sm->getCamera("PlayerCam");
		mTopCam = sm->getCamera("TopCam");
		mSceneMgr = sm;
        mWindow = win;
		mRotateSpeed = rs;
        mMoveSpeed = ms;
		mRotationPeriod = rp;
		mDeathAngle = 0;
        mStatsOn = true;
		mNumScreenShots = 0;
		mTimeUntilNextToggle = 0;
        mSceneDetailIndex = 0;
        mMoveScale = 0.0f;
        mRotScale = 0.0f;
	    mTranslateVector = Vector3::ZERO;
        mAniso = 8;
        mFiltering = TFO_ANISOTROPIC;

        MaterialManager::getSingleton().setDefaultTextureFiltering(mFiltering);
        MaterialManager::getSingleton().setDefaultAnisotropy(mAniso);

		initKdTreeOverlay();

        showDebugOverlay(true);
    }
    virtual ~KdTreeAppListener()
    {
		if (mInputTypeSwitchingOn)
		{
            delete mEventProcessor;
		}
        else
        {
            PlatformManager::getSingleton().destroyInputReader( mInputDevice );
        }
    }

	void initKdTreeOverlay()
	{
		String na = " N/A";
		String value;

		int maxd;
		if (mSceneMgr->getOption("KdTreeMaxDepth", &maxd))
			value = " " + StringConverter::toString(maxd);
		else
			value = na;
		String caption = OverlayManager::getSingleton().getOverlayElement("KdTree/Depth")->getCaption();
		OverlayManager::getSingleton().getOverlayElement("KdTree/Depth")->setCaption(caption + value);

		Real kt;
		if (mSceneMgr->getOption("KT", &kt))
			value = " " + StringConverter::toString(kt);
		else
			value = na;
		caption = OverlayManager::getSingleton().getOverlayElement("KdTree/KT")->getCaption();
		OverlayManager::getSingleton().getOverlayElement("KdTree/KT")->setCaption(caption + value);

		Real ki;
		if (mSceneMgr->getOption("KI", &ki))
			value = " " + StringConverter::toString(ki);
		else
			value = na;
		caption = OverlayManager::getSingleton().getOverlayElement("KdTree/KI")->getCaption();
		OverlayManager::getSingleton().getOverlayElement("KdTree/KI")->setCaption(caption + value);

		int hl;
		if (mSceneMgr->getOption("HighlightLevel", &hl))
			value = " " + StringConverter::toString(hl);
		else
			value = na;
		if (!mShowTree)
			value = " off";
		caption = OverlayManager::getSingleton().getOverlayElement("KdTree/Highlight")->getCaption();
		OverlayManager::getSingleton().getOverlayElement("KdTree/Highlight")->setCaption(caption + value);

		String bm;
		if (mSceneMgr->getOption("BuildMethod", &bm))
			value = " " + bm;
		else
			value = na;
		caption = OverlayManager::getSingleton().getOverlayElement("KdTree/BuildMethod")->getCaption();
		OverlayManager::getSingleton().getOverlayElement("KdTree/BuildMethod")->setCaption(caption + value);

		String rm;
		if (mSceneMgr->getOption("RenderMethod", &rm))
			value = " " + rm;
		else
			value = na;
		caption = OverlayManager::getSingleton().getOverlayElement("KdTree/RenderMethod")->getCaption();
		OverlayManager::getSingleton().getOverlayElement("KdTree/RenderMethod")->setCaption(caption + value);

		value = na;
		if (mFreeMove)
			value = " Free";
		else
			value = " Ground";
		caption = OverlayManager::getSingleton().getOverlayElement("KdTree/Movement")->getCaption();
		OverlayManager::getSingleton().getOverlayElement("KdTree/Movement")->setCaption(caption + value);
	}

	void toggleCullCamera()
	{
		mCullCamera = !mCullCamera;
		int zorder = 2; 
		if (mCullCamera)
		{
			Viewport* tvp = mWindow->addViewport(mTopCam,zorder,0.66,0.66,0.34,0.34);
			tvp->setBackgroundColour(ColourValue(1.0, 0.0, 0.0));
			tvp->setOverlaysEnabled(false);

			mTopCam->setAspectRatio(
				Real(tvp->getActualWidth())/Real(tvp->getActualHeight()));
		}
		else
		{
			mWindow->removeViewport(zorder);
		}
	}

    virtual bool processUnbufferedKeyInput(const FrameEvent& evt)
    {
		// moving the cull camera
		if (mInputDevice->isKeyDown(KC_NUMPAD2))
		{
			mCullTransVec.y = -mMoveScale;
		}
		if (mInputDevice->isKeyDown(KC_NUMPAD8))
		{
			mCullTransVec.y = mMoveScale;
		}
		if (mInputDevice->isKeyDown(KC_NUMPAD4))
		{
			mCullTransVec.x = mMoveScale;
		}
		if (mInputDevice->isKeyDown(KC_NUMPAD6))
		{
			mCullTransVec.x = -mMoveScale;
		}
		if (mInputDevice->isKeyDown(KC_SUBTRACT))
		{
			mCullTransVec.z = mMoveScale;
		}
		if (mInputDevice->isKeyDown(KC_ADD))
		{
			mCullTransVec.z = -mMoveScale;
		}

		// stuff for the kdtree overlay
		static String sMaxDepth = "[1][2] Max Tree Depth: ";
		static String sHL = "[3][4] Highlight: ";
		static String sKT = "[5][6] KT: ";
		static String sKI = "[7][8] KI: ";
		static String sBM = "[Y] Build Method: ";
		static String sRM = "[X] Render Method: ";
		static String sMov = "[Q] Movement: ";

		static String bmOptions[] = { "Recursive", "PriorityQueue" };
		static int bmOptionsSize = sizeof(bmOptions) / sizeof(String);
		static int bmCurrent = 0;

		//static String rmOptions[] = { "Recursive", "Stack", "StopAndWait", "CHC" };
		static String rmOptions[] = { "Recursive", "Stack" };
		static int rmOptionsSize = sizeof(rmOptions) / sizeof (String);
		static int rmCurrent = 0;

		if ((mInputDevice->isKeyDown(KC_1) || mInputDevice->isKeyDown(KC_2)) && mTimeUntilNextToggle <= 0)
		{
			int currdepth;
			if (mSceneMgr->getOption("KdTreeMaxDepth", &currdepth))
			{
				if (mInputDevice->isKeyDown(KC_1))
					currdepth--;
				else if (mInputDevice->isKeyDown(KC_2))
					currdepth++;

				if (mSceneMgr->setOption("KdTreeMaxDepth", &currdepth))
				{
					OverlayManager::getSingleton().getOverlayElement("KdTree/Depth")->setCaption(sMaxDepth + StringConverter::toString(currdepth));
					int hl;
					if (mSceneMgr->getOption("HighlightLevel", &hl))
						OverlayManager::getSingleton().getOverlayElement("KdTree/Highlight")->setCaption(sHL + StringConverter::toString(hl));
					else
						OverlayManager::getSingleton().getOverlayElement("KdTree/Highlight")->setCaption(sHL + "N/A");
				}
			}
			else
			{
				OverlayManager::getSingleton().getOverlayElement("KdTree/Depth")->setCaption(sMaxDepth + "N/A");
				OverlayManager::getSingleton().getOverlayElement("KdTree/Highlight")->setCaption(sHL + "N/A");
			}

			if (!mShowTree)
				OverlayManager::getSingleton().getOverlayElement("KdTree/Highlight")->setCaption(sHL + "off");

			mTimeUntilNextToggle = 0.2;
		}

		if ((mInputDevice->isKeyDown(KC_3) || mInputDevice->isKeyDown(KC_4)) && mTimeUntilNextToggle <= 0)
		{
			int hl;
			if (mSceneMgr->getOption("HighlightLevel", &hl))
			{
				if (mInputDevice->isKeyDown(KC_3))
					hl--;
				else if (mInputDevice->isKeyDown(KC_4))
					hl++;

				if (mSceneMgr->setOption("HighlightLevel", &hl))
					OverlayManager::getSingleton().getOverlayElement("KdTree/Highlight")->setCaption(sHL + StringConverter::toString(hl));
			}
			else
			{
				OverlayManager::getSingleton().getOverlayElement("KdTree/Highlight")->setCaption(sHL + "N/A");
			}

			if (!mShowTree)
				OverlayManager::getSingleton().getOverlayElement("KdTree/Highlight")->setCaption(sHL + "off");

			mTimeUntilNextToggle = 0.2;
		}


		if ((mInputDevice->isKeyDown(KC_5) ||mInputDevice->isKeyDown(KC_6)) && mTimeUntilNextToggle <= 0)
		{
			Real kt;
			if (mSceneMgr->getOption("KT", &kt))
			{
				if (mInputDevice->isKeyDown(KC_5))
					kt -= 0.1;
				else if (mInputDevice->isKeyDown(KC_6))
					kt += 0.1;

				if (kt < 0.1)
					kt = 0.1;
			
				if (mSceneMgr->setOption("KT", &kt))
					OverlayManager::getSingleton().getOverlayElement("KdTree/KT")->setCaption(sKT + StringConverter::toString(kt));
			}
			else
			{
				OverlayManager::getSingleton().getOverlayElement("KdTree/KT")->setCaption(sKT + "N/A");
			}

			mTimeUntilNextToggle = 0.2;
		}

		if ((mInputDevice->isKeyDown(KC_7) || mInputDevice->isKeyDown(KC_8)) && mTimeUntilNextToggle <= 0)
		{
			Real ki;
			if (mSceneMgr->getOption("KI", &ki))
			{
				if (mInputDevice->isKeyDown(KC_7))
					ki -= 0.1;
				else if (mInputDevice->isKeyDown(KC_8))
					ki += 0.1;

				if (ki < 0.1)
					ki = 0.1;

				if (mSceneMgr->setOption("KI", &ki))
					OverlayManager::getSingleton().getOverlayElement("KdTree/KI")->setCaption(sKI + StringConverter::toString(ki));
			}
			else
			{
				OverlayManager::getSingleton().getOverlayElement("KdTree/KI")->setCaption(sKI + "N/A");
			}

			mTimeUntilNextToggle = 0.2;
		}

		if (mInputDevice->isKeyDown(KC_Z) && mTimeUntilNextToggle <= 0)
		{
			String bm;
			if (mSceneMgr->getOption("BuildMethod", &bm))
			{
				bmCurrent = (bmCurrent + 1) % bmOptionsSize;
				if (mSceneMgr->setOption("BuildMethod", &bmOptions[bmCurrent]))
					OverlayManager::getSingleton().getOverlayElement("KdTree/BuildMethod")->setCaption(sBM + bmOptions[bmCurrent]);
			}
			else
			{
				OverlayManager::getSingleton().getOverlayElement("KdTree/BuildMethod")->setCaption(sBM + "N/A");
			}

			mTimeUntilNextToggle = 0.5;
		}

		if (mInputDevice->isKeyDown(KC_X) && mTimeUntilNextToggle <= 0)
		{
			String rm;
			if (mSceneMgr->getOption("RenderMethod", &rm))
			{
				rmCurrent = (rmCurrent + 1) % rmOptionsSize;
				if (mSceneMgr->setOption("RenderMethod", &rmOptions[rmCurrent]))
					OverlayManager::getSingleton().getOverlayElement("KdTree/RenderMethod")->setCaption(sRM + rmOptions[rmCurrent]);
			}
			else
			{
				OverlayManager::getSingleton().getOverlayElement("KdTree/RenderMethod")->setCaption(sRM + "N/A");
			}

			mTimeUntilNextToggle = 0.5;
		}

		if (mInputDevice->isKeyDown(KC_Q) && mTimeUntilNextToggle <= 0)
		{
			mFreeMove = !mFreeMove;
			String move = "N/A";
			if (mFreeMove)
				move = "Free";
			else
				move = "Ground";

			OverlayManager::getSingleton().getOverlayElement("KdTree/Movement")->setCaption(sMov + move);

			mTimeUntilNextToggle = 0.5;
		}


		if (mInputDevice->isKeyDown(KC_BACK) && mTimeUntilNextToggle <= 0)
		{
			mSceneMgr->setOption("RebuildKdTree", 0);
			mTimeUntilNextToggle = 1;
		}

		if (mInputDevice->isKeyDown(KC_N) && mTimeUntilNextToggle <= 0)
		{
			//Entity * ent = mSceneMgr->createEntity("randominsert" + StringConverter::toString(mSeqNum), "robot.mesh");
			Entity * ent = mSceneMgr->createEntity("randominsert" + StringConverter::toString(mSeqNum), "razor.mesh");
			//Vector3 position(Math::RangeRandom(100, 1125), -0, Math::RangeRandom(-1125, 1125));
			Vector3 position(Math::RangeRandom(-5000, 5000), Math::RangeRandom(-5000, 5000), Math::RangeRandom(-5000, 5000));
			//Quaternion orientation(Radian(Math::RangeRandom(-Math::PI, Math::PI)), Vector3::UNIT_Y);
			Vector3 axis(Math::RangeRandom(-1,1),Math::RangeRandom(-1,1),Math::RangeRandom(-1,1));
			axis.normalise();
			Quaternion orientation(Radian(Math::RangeRandom(-Math::PI, Math::PI)), axis);
			Vector3 scale(Math::RangeRandom(0.5, 5),Math::RangeRandom(0.5, 5),Math::RangeRandom(0.5, 5));
			SceneNode * anchor = mSceneMgr->getSceneNode("AnchorNode");
			SceneNode *sn = anchor->createChildSceneNode("RandomInsertNode" + StringConverter::toString(mSeqNum), position, orientation);
			sn->attachObject(ent);
			sn->setScale(scale);
			mTimeUntilNextToggle = 0.5;
			mSeqNum++;
		}

		if (mInputDevice->isKeyDown(KC_J) && mTimeUntilNextToggle <= 0)
		{
			if (mSeqNum > 0)
			{
				mSeqNum--;
				mSceneMgr->destroySceneNode("RandomInsertNode" + StringConverter::toString(mSeqNum));
				mSceneMgr->destroyEntity("randominsert" + StringConverter::toString(mSeqNum));
				mTimeUntilNextToggle = 0.5;
			}
		}

		if (mInputDevice->isKeyDown(KC_Y) && mTimeUntilNextToggle <= 0)
		{
			LogManager::getSingleton().logMessage("############## Camera Position:");
			LogManager::getSingleton().logMessage("############## " + StringConverter::toString(mCamera->getPosition()));
			LogManager::getSingleton().logMessage("############## " + StringConverter::toString(mCamera->getOrientation()));
			LogManager::getSingleton().logMessage("############## Cull Camera Position:");
			LogManager::getSingleton().logMessage("############## " + StringConverter::toString(mTopCam->getPosition()));
			LogManager::getSingleton().logMessage("############## " + StringConverter::toString(mTopCam->getOrientation()));
			mTimeUntilNextToggle = 1.0;
		}

		if (mInputDevice->isKeyDown(KC_B) && mTimeUntilNextToggle <= 0)
		{
			mSceneMgr->showBoundingBoxes( ! mSceneMgr->getShowBoundingBoxes() );
			mTimeUntilNextToggle = 0.5;
		}

		if (mInputDevice->isKeyDown(KC_C) && mTimeUntilNextToggle <= 0)
		{
			toggleCullCamera();
			mTimeUntilNextToggle = 0.5;
		}

		if (mInputDevice->isKeyDown(KC_V) && mTimeUntilNextToggle <= 0)
		{
			mShowTree = !mShowTree;

			if (mSceneMgr->getTypeName() == "OctreeSceneManager")
			{
				//(static_cast<OctreeSceneManager *>(mSceneMgr))->setShowBoxes(mShowTree);
				mSceneMgr->setOption("ShowOctree", &mShowTree);
			}
			else if (mSceneMgr->getTypeName() == "KdTreeSceneManager")
			{
				mSceneMgr->setOption("ShowKdTree", &mShowTree);
				//KdTreeSceneManager *sm = static_cast<KdTreeSceneManager *>(mSceneMgr);
				//sm->setShowBoxes(!sm->getShowBoxes());
				//(static_cast<KdTreeSceneManager *>(mSceneMgr))->setShowBoxes(mShowTree);
			}

			if (!mShowTree)
				OverlayManager::getSingleton().getOverlayElement("KdTree/Highlight")->setCaption(sHL + "off");
			else
			{
				int hl;
				if (mSceneMgr->getOption("HighlightLevel", &hl))
					OverlayManager::getSingleton().getOverlayElement("KdTree/Highlight")->setCaption(sHL + StringConverter::toString(hl));
				else
					OverlayManager::getSingleton().getOverlayElement("KdTree/Highlight")->setCaption(sHL + "N/A");
			}
			
			mTimeUntilNextToggle = 0.5;
		}

		if (mInputDevice->isKeyDown(KC_G) && mTimeUntilNextToggle <= 0)
		{
			bool toggleShowAllBoxes;
			mSceneMgr->getOption("ShowAllBoxes", &toggleShowAllBoxes);
			toggleShowAllBoxes = !toggleShowAllBoxes;
			mSceneMgr->setOption("ShowAllBoxes", &toggleShowAllBoxes);
			mTimeUntilNextToggle = 0.2;
		}

		if (mInputDevice->isKeyDown(KC_P))
		{
			mWindow->writeContentsToTimestampedFile("Screenshot",".jpg");
		}

        if (mInputDevice->isKeyDown(KC_A))
        {
            // Move camera left
            mTranslateVector.x = -mMoveScale;
        }

        if (mInputDevice->isKeyDown(KC_D))
        {
            // Move camera RIGHT
            mTranslateVector.x = mMoveScale;
        }

        /* Move camera forward by keypress. */
        if (mInputDevice->isKeyDown(KC_UP) || mInputDevice->isKeyDown(KC_W) )
        {
            mTranslateVector.z = -mMoveScale;
        }

        /* Move camera backward by keypress. */
        if (mInputDevice->isKeyDown(KC_DOWN) || mInputDevice->isKeyDown(KC_S) )
        {
            mTranslateVector.z = mMoveScale;
        }

        if (mInputDevice->isKeyDown(KC_PGUP))
        {
            // Move camera up
            mTranslateVector.y = mMoveScale;
        }

        if (mInputDevice->isKeyDown(KC_PGDOWN))
        {
            // Move camera down
            mTranslateVector.y = -mMoveScale;
        }

        if (mInputDevice->isKeyDown(KC_RIGHT))
        {
            mCamera->yaw(-mRotScale);
        }
		
        if (mInputDevice->isKeyDown(KC_LEFT))
        {
            mCamera->yaw(mRotScale);
        }

        if( mInputDevice->isKeyDown( KC_ESCAPE) )
        {            
            return false;
        }

		// see if switching is on, and you want to toggle 
        if (mInputTypeSwitchingOn && mInputDevice->isKeyDown(KC_M) && mTimeUntilNextToggle <= 0)
        {
			switchMouseMode();
            mTimeUntilNextToggle = 1;
        }

        if (mInputTypeSwitchingOn && mInputDevice->isKeyDown(KC_K) && mTimeUntilNextToggle <= 0)
        {
			// must be going from immediate keyboard to buffered keyboard
			switchKeyMode();
            mTimeUntilNextToggle = 1;
        }
        if (mInputDevice->isKeyDown(KC_F) && mTimeUntilNextToggle <= 0)
        {
            mStatsOn = !mStatsOn;
            showDebugOverlay(mStatsOn);

            mTimeUntilNextToggle = 1;
        }
        if (mInputDevice->isKeyDown(KC_T) && mTimeUntilNextToggle <= 0)
        {
            switch(mFiltering)
            {
            case TFO_BILINEAR:
                mFiltering = TFO_TRILINEAR;
                mAniso = 1;
                break;
            case TFO_TRILINEAR:
                mFiltering = TFO_ANISOTROPIC;
                mAniso = 8;
                break;
            case TFO_ANISOTROPIC:
                mFiltering = TFO_BILINEAR;
                mAniso = 1;
                break;
            default:
                break;
            }
            MaterialManager::getSingleton().setDefaultTextureFiltering(mFiltering);
            MaterialManager::getSingleton().setDefaultAnisotropy(mAniso);


            showDebugOverlay(mStatsOn);

            mTimeUntilNextToggle = 1;
        }

        if (mInputDevice->isKeyDown(KC_SYSRQ) && mTimeUntilNextToggle <= 0)
        {
			char tmp[20];
			sprintf(tmp, "screenshot_%d.png", ++mNumScreenShots);
            mWindow->writeContentsToFile(tmp);
            mTimeUntilNextToggle = 0.5;
			mWindow->setDebugText(String("Wrote ") + tmp);
        }
		
		if (mInputDevice->isKeyDown(KC_R) && mTimeUntilNextToggle <=0)
		{
			mSceneDetailIndex = (mSceneDetailIndex+1)%3 ;
			switch(mSceneDetailIndex) {
				case 0 : mCamera->setPolygonMode(PM_SOLID) ; break ;
				case 1 : mCamera->setPolygonMode(PM_WIREFRAME) ; break ;
				case 2 : mCamera->setPolygonMode(PM_POINTS) ; break ;
			}
			mTimeUntilNextToggle = 0.5;
		}

        static bool displayCameraDetails = false;
        if (mInputDevice->isKeyDown(KC_P) && mTimeUntilNextToggle <= 0)
        {
            displayCameraDetails = !displayCameraDetails;
            mTimeUntilNextToggle = 0.5;
            if (!displayCameraDetails)
                mWindow->setDebugText("");
        }
        if (displayCameraDetails)
        {
            // Print camera details
            mWindow->setDebugText("P: " + StringConverter::toString(mCamera->getDerivedPosition()) + " " + 
                "O: " + StringConverter::toString(mCamera->getDerivedOrientation()));
        }

        // Return true to continue rendering
        return true;
    }

    bool processUnbufferedMouseInput(const FrameEvent& evt)
    {
        /* Rotation factors, may not be used if the second mouse button is pressed. */

        /* If the second mouse button is pressed, then the mouse movement results in 
           sliding the camera, otherwise we rotate. */
        if( mInputDevice->getMouseButton( 1 ) )
        {
            mTranslateVector.x += mInputDevice->getMouseRelativeX() * 0.13;
            mTranslateVector.y -= mInputDevice->getMouseRelativeY() * 0.13;
        }
        else
        {
            mRotX = Degree(-mInputDevice->getMouseRelativeX() * 0.13);
            mRotY = Degree(-mInputDevice->getMouseRelativeY() * 0.13);
        }


		return true;
	}

	void moveCamera()
	{

        // Make all the changes to the camera
        // Note that YAW direction is around a fixed axis (freelook style) rather than a natural YAW (e.g. airplane)
        mCamera->yaw(mRotX);
        mCamera->pitch(mRotY);
        mCamera->moveRelative(mTranslateVector);
		mTopCam->moveRelative(mCullTransVec);

		if (!mFreeMove)
		{
			Vector3 pos = mCamera->getPosition();
			RaySceneQuery * rsquery = mSceneMgr->createRayQuery(Ray(pos, Vector3::NEGATIVE_UNIT_Y));
			RaySceneQueryResult& rsresult = rsquery->execute();

			RaySceneQueryResult::iterator it = rsresult.begin();
			RaySceneQueryResult::iterator end = rsresult.end();

			while (it != end && it->movable)
			{
				if (it->movable->getName() != "PlayerCam")
				{
					mCamera->setPosition(pos.x, it->movable->getWorldBoundingBox().getCenter().y + 3, pos.z);
					break;
				}
				it++;
			}
			OGRE_DELETE(rsquery);
		}
	}

    void showDebugOverlay(bool show)
    {
        if (mDebugOverlay)
        {
            if (show)
            {
                mDebugOverlay->show();
				mKdTreeOverlay->show();
            }
            else
            {
                mDebugOverlay->hide();
				mKdTreeOverlay->hide();
            }
        }
    }

    // Override frameStarted event to process that (don't care about frameEnded)
    bool frameStarted(const FrameEvent& evt)
    {
        if(mWindow->isClosed())
            return false;

        if (!mInputTypeSwitchingOn)
    	{
            mInputDevice->capture();
        }


		if ( !mUseBufferedInputMouse || !mUseBufferedInputKeys)
		{
			// one of the input modes is immediate, so setup what is needed for immediate mouse/key movement
			if (mTimeUntilNextToggle >= 0) 
				mTimeUntilNextToggle -= evt.timeSinceLastFrame;

			// If this is the first frame, pick a speed
			if (evt.timeSinceLastFrame == 0)
			{
				mMoveScale = 1;
				mRotScale = 0.1;
			}
			// Otherwise scale movement units by time passed since last frame
			else
			{
				// Move about 100 units per second,
				mMoveScale = mMoveSpeed * evt.timeSinceLastFrame;
				// Take about 10 seconds for full rotation
				mRotScale = mRotateSpeed * evt.timeSinceLastFrame;
			}
			mRotX = 0;
            mRotY = 0;
	        mTranslateVector = Vector3::ZERO;
			mCullTransVec = Vector3::ZERO;
		}

        if (mUseBufferedInputKeys)
        {
            // no need to do any processing here, it is handled by event processor and 
			// you get the results as KeyEvents
        }
        else
        {
            if (processUnbufferedKeyInput(evt) == false)
			{
				return false;
			}
        }
        if (mUseBufferedInputMouse)
        {
            // no need to do any processing here, it is handled by event processor and 
			// you get the results as MouseEvents
        }
        else
        {
            if (processUnbufferedMouseInput(evt) == false)
			{
				return false;
			}
        }

		if ( !mUseBufferedInputMouse || !mUseBufferedInputKeys)
		{
			// one of the input modes is immediate, so update the movement vector

			moveCamera();

		}

		Camera	*followCam = mSceneMgr->getCamera("FollowCam");
		if (followCam->getAutoTrackTarget() != 0)
		{
			// Move the death thingy & update lookat camera FOV
			SceneNode *deathPivotNode = mSceneMgr->getSceneNode("deathPivotNode");
			SceneNode *deathNode = mSceneMgr->getSceneNode("movingNode");
			SceneNode *pivotNode = mSceneMgr->getSceneNode("pivotNode");

			Vector3 fcpos = followCam->getPosition();
			Vector3 oldpos = deathNode->getWorldPosition();
			
			Radian deltaAngle = Radian(evt.timeSinceLastFrame / mRotationPeriod * Math::TWO_PI);
			deathPivotNode->rotate(Vector3::UNIT_Y, deltaAngle);
			pivotNode->rotate(Vector3::UNIT_Y, deltaAngle * 4);
			
			Vector3 pos = deathNode->getWorldPosition();

			Real olddist = (oldpos - fcpos).length();
			Real dist = (pos - fcpos).length();

			Radian oldfov = followCam->getFOVy();
			Radian fov = Radian(olddist / dist) * oldfov;

			followCam->setFOVy(fov);
		}

		return true;
    }

    bool frameEnded(const FrameEvent& evt)
    {
        updateStats();
        return true;
    }

	void switchMouseMode() 
	{
        mUseBufferedInputMouse = !mUseBufferedInputMouse;
		mInputDevice->setBufferedInput(mUseBufferedInputKeys, mUseBufferedInputMouse);
	}
	void switchKeyMode() 
	{
        mUseBufferedInputKeys = !mUseBufferedInputKeys;
		mInputDevice->setBufferedInput(mUseBufferedInputKeys, mUseBufferedInputMouse);
	}

	void keyClicked(KeyEvent* e) 
	{
		if (e->getKeyChar() == 'm')
		{
			switchMouseMode();
		}
		else if (e->getKeyChar() == 'k')
		{

			switchKeyMode();
		}

	}
	void keyPressed(KeyEvent* e) {}
	void keyReleased(KeyEvent* e) {}

protected:
    EventProcessor* mEventProcessor;
    InputReader* mInputDevice;
    Camera* mCamera;
	Camera* mTopCam;
	SceneManager *mSceneMgr;
	Radian mDeathAngle;

    Vector3 mTranslateVector;
    RenderWindow* mWindow;
    bool mStatsOn;
    bool mUseBufferedInputKeys, mUseBufferedInputMouse, mInputTypeSwitchingOn;
	unsigned int mNumScreenShots;
    float mMoveScale;
    Degree mRotScale;
    // just to stop toggles flipping too fast
    Real mTimeUntilNextToggle ;
    Radian mRotX, mRotY;
    TextureFilterOptions mFiltering;
    int mAniso;

	// stuff for the culling camera
	bool mCullCamera;
	Vector3 mCullTransVec;
};

#endif

